/* -----------------------------------------------------------------------------
 *
 *	bitloop~
 *		a variable-size circular buffer
 *		
 *
 *	2007 . Cooper Baker . www.cooperbaker.com
 *
 *----------------------------------------------------------------------------*/

#include <Carbon/Carbon.h>
#include <MaxAPI/MaxAPI.h>
#include <MaxAudioAPI/MaxAudioAPI.h>

void *bitloop_class;		// pointer to this 'class'

// bitloop~ structure ------------------------------------------------------ 
typedef struct _bitloop	// bitloop data structure
{ 
    t_pxobject x_obj;		// every msp external needs this; t_pxobject struct contains object info
	float	sample_rate;	// sample rate 
	long	write_samp;	// input sample
	long	read_samp;	// output sample
	float	length_msec;	// delay length in milliseconds
	long	length_samps;	// delay length in samples
	long	buffer_samps;	// buffer length in samples
	float	buffer_msec;	// buffer length in milliseconds
	float	in1_conn;		// inlet one connected flag
	float	*buffer;	// pointer to the delay buffer
} t_bitloop; 

// prototypes ------------------------------------------------------------------ 
void *bitloop_new(double maximum);				// makes a new bitloop~ object
void bitloop_free(t_bitloop *x);				// deletes the bitloop~ object
void bitloop_dsp(t_bitloop *x, t_signal **sp, short *count);	// sets up dsp for bitloop
t_int *bitloop_perform(t_int *w);				// the dsp perform function
void bitloop_assist(t_bitloop *x, void *b, long m, long a, char *s);// inlet / outlet assistance

// main ------------------------------------------------------------------------
void main(void) 
{ 
	setup((t_messlist **)&bitloop_class, (method)bitloop_new, (method)bitloop_free, (short)sizeof(t_bitloop), 0L, A_DEFFLOAT, 0); 
	addmess((method)bitloop_dsp, "dsp", A_CANT, 0);	// add dsp method
	addmess((method)bitloop_assist, "assist", A_CANT, 0);	// assist method 
	dsp_initclass();					// init the class
	post( "" );								
	post( "cb.bitloop~ :: variable length delay line" );
	post( "2007 . Cooper Baker . www.cooperbaker.com" );
	post( "" );
}								// post after init completes

// bitloop_dsp ----------------------------------------------------------------
void bitloop_dsp(t_bitloop *x, t_signal **sp, short *count)
{				// bitloop_dsp(*data struct, **signal struct, signal count )
	x->in1_conn = count[0] && count[1];
	x->sample_rate = sp[0]->s_sr;
	x->buffer_samps = ( ( x->buffer_msec * 0.001 ) * sp[0]->s_sr );
	dsp_add( bitloop_perform, 6, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[0]->s_n, count, x );
}				// 6 = number of values passed to bitloop_perform

// bitloop_perform ------------------------------------------------------------ 
t_int *bitloop_perform(t_int *w)				// bitloop_perform runs once per signal vector
{ 
	t_float *in1	= (t_float *)(w[1]);			// left inlet signal vector location
	t_float *in2	= (t_float *)(w[2]);			// right inlet signal vector location
	t_float *out	= (t_float *)(w[3]);			// outlet signal vector location
	int n			= (int)(w[4]);			// vector size in samples
	t_bitloop *x	= (t_bitloop *)(w[5]);			// bitloop struct location
	float			sample;				// for the current sample
	float			dtime;
	
// DSP BLOCK ///////////////////////////////////////////////////////////////////
	while ( n-- )						// while a sample remains, do dsp ( n-- is the counter )
	{  		
		sample = *in1++;				// audio signal in
		dtime = *in2++;					// delay time signal in
		
		if( x->in1_conn )
			x->length_msec = dtime;			// msec delay time signal in
		else
			x->length_msec = x->buffer_msec;
			
		x->length_samps = ( ( x->length_msec * 0.001 ) * x->sample_rate );
		
		x->length_samps = ( x->length_samps > 0 ) ? x->length_samps : 1 ;
		
								// calculate delay sample length
		*( x->buffer + x->write_samp ) = sample;	// write to buffer via pointer arithmetic
		x->read_samp = x->write_samp - x->length_samps;// calculate read location based on write location
		if( x->read_samp < 0 ) x->read_samp = x->buffer_samps + x->read_samp; // wrap read_samp
		sample = *( x->buffer + x->read_samp );	// read from buffer via pointer arithemtic
		
		x->write_samp++;				// increment write sample position
		if( x->write_samp >= x->buffer_samps ) x->write_samp = x->write_samp - x->buffer_samps ; // >= !!!
								// wrap write_samp
		*out++ = sample;				// audio sample out
	}
// END DSP BLOCK ///////////////////////////////////////////////////////////////	
	
	return (w + 7);		// always return w plus one more than the 2nd argument in dsp_add() 
} 

// bitloop_assist -------------------------------------------------------------
void bitloop_assist(t_bitloop *x, void *b, long m, long a, char *s)
{
	switch (m)						// inlet / outlet switch
	{
		case 1:						// 1 = inlets
			switch (a)				// inlet number switch
			{
				case 0:				// inlet 0
					sprintf(s, "(signal) Audio");
					break;			// assist message
				case 1:				// inlet 1
					sprintf(s, "(signal) Milliseconds");
					break;			// assist message
				default:
					break;
			}
			break;
		case 2:						// 2 = outlets
			switch (a)				// outlet number switch
			{
				case 0:				// outlet 0
					sprintf(s, "(signal) Audio");
					break;			// assist message
				default:
					break;
			}
			break;
	}
}

// bitloop_new() -------------------------------------------------------------- 
void *bitloop_new(double maximum)				// bitloop~ 'constructor'
{								// register the object with msp and save its address as *x
	t_bitloop *x = (t_bitloop *)newobject(bitloop_class);
	dsp_setup((t_pxobject *)x, 2);				// create inlets
	outlet_new((t_pxobject *)x, "signal");			// create outlet 

	x->buffer_msec		= ( maximum > 0 ) ? maximum : 1000 ;
	x->write_samp	= 1;					// init write_samp value
	x->length_msec	= 0.1;					// init length_msec value
	
	x->buffer = (float *)sysmem_newptrclear( (maximum * 0.001 ) * 192000 );
								// allocate enough memory for 192kHz
	return (x);						// where am i?
} 

// bitloop_free() -----------------------------------------------------------------
void bitloop_free(t_bitloop *x)				// bitloop~ 'destructor'
{

	sysmem_freeptr( x->buffer );				// collect the garbage
	dsp_free((t_pxobject *)x);				// msp free function
}