/* -----------------------------------------------------------------------------
 *
 *	dropper~
 *		causes a single-sample drop-out at each wave peak
 *		
 *
 *	2007 . Cooper Baker . www.cooperbaker.com
 *
 *----------------------------------------------------------------------------*/

#include <Carbon/Carbon.h>
#include <MaxAPI/MaxAPI.h>
#include <MaxAudioAPI/MaxAudioAPI.h>

void *dropper_class;									// pointer to this 'class'

// dropper~ structure ------------------------------------------------------ 
typedef struct _dropper									// dropper data structure
{ 
    t_pxobject x_obj;									// every max external needs this; t_pxobject struct contains object info
	Float32		sample[3];								// samples: x[n+1] x[n] x[n-1]
	Float32		drop;
} t_dropper; 

// prototypes ------------------------------------------------------------------ 
void *dropper_new(void);								// makes a new dropper~ object
void dropper_free(t_dropper *x);						// deletes the dropper~ object
void dropper_dsp(t_dropper *x, t_signal **sp, short *count); // sets up dsp for dropper
t_int *dropper_perform(t_int *w);						// the dsp perform function
void dropper_assist(t_dropper *x, void *b, long m, long a, char *s);
														// inlet / outlet assistance
// main ------------------------------------------------------------------------
void main(void) 
{ 
	setup((t_messlist **)&dropper_class, (method)dropper_new, (method)dropper_free, (short)sizeof(t_dropper), 0L, 0); 
	addmess((method)dropper_dsp, "dsp", A_CANT, 0);		// add dsp method
	addmess((method)dropper_assist, "assist", A_CANT, 0);// assist method 
	dsp_initclass();	
	post( "" );								
	post( "cb.dropper~ :: causes a single-sample drop-out at each wave peak" );
	post( "2007 . Cooper Baker . www.cooperbaker.com" );
	post( "" );
}														// post after init completes

// dropper_dsp --------------------------------------------------------------------
void dropper_dsp(t_dropper *x, t_signal **sp, short *count)	// dropper_dsp(*data struct, **signal struct, signal count )
{ 
	dsp_add( dropper_perform, 4, sp[0]->s_vec, sp[1]->s_vec, sp[0]->s_n, x );
}						   // 4 = number of values passed to dropper_perform

// dropper_perform ---------------------------------------------------------------- 
t_int *dropper_perform(t_int *w)						// dropper_perform runs once per signal vector
{ 
	t_float *in	= (t_float *)(w[1]);					// inlet signal vector location
	t_float *out	= (t_float *)(w[2]);				// outlet signal vector location
	int n		= (int)(w[3]);							// vector size in samples
	t_dropper *x	= (t_dropper *)(w[4]);				// dropper struct location
	
	int i;
// DSP BLOCK ///////////////////////////////////////////////////////////////////	
	while ( n-- )										// while a sample remains, do dsp ( n-- is the counter )
	{  
		x->sample[2] = x->sample[1];
		x->sample[1] = x->sample[0];
		x->sample[0] = *in++;							// *in++/*out++ should only be done once per while loop cycle
		
		x->drop = x->sample[1];
		
		if( x->sample[1] > 0 )
		{
			if( ( x->sample[0] < x->sample[1] ) && ( x->sample[1] > x->sample[2] ) )
			{
				x->drop = 0;
			}
		}
		if( x->sample[1] < 0 )
		{
			if( ( x->sample[0] > x->sample[1] ) && ( x->sample[1] < x->sample[2] ) )
			{
				x->drop = 0;
			}
		}
		*out++ = x->drop;								// sample is used to make things clearer
	}
// END DSP BLOCK ///////////////////////////////////////////////////////////////	
	
	return (w + 5);										// always return w plus one more than the 2nd argument in dsp_add() 
} 

// dropper_assist -----------------------------------------------------------------
void dropper_assist(t_dropper *x, void *b, long m, long a, char *s)
{
	switch (m)											// inlet / outlet switch
	{
		case 1:											// 1 = inlets
			switch (a)									// inlet number switch
			{
				case 0:									// inlet 0
					sprintf(s, "(signal) Thru~ Input");
					break;								// assist message
				default:
					break;
			}
			break;
		case 2:											// 2 = outlets
			switch (a)									// outlet number switch
			{
				case 0:									// outlet 0
					sprintf(s, "(signal) Thru~ Output");
					break;								// assist message
				default:
					break;
			}
			break;
	}
}

// dropper_new() ------------------------------------------------------------------ 
void *dropper_new(void)									// dropper~ 'constructor'
{ 
	t_dropper *x = (t_dropper *)newobject(dropper_class);// register the object with dropper and save its address as *x
	dsp_setup((t_pxobject *)x, 1);						// create left inlet 
	outlet_new((t_pxobject *)x, "signal");				// create left outlet 
	
	x->sample[0] = 0;									// init x->samples
	x->sample[1] = 0;
	x->sample[2] = 0;
	
	return (x);											// where am i?
} 

// dropper_free() -----------------------------------------------------------------
void dropper_free(t_dropper *x)							// dropper~ 'destructor'
{
	dsp_free((t_pxobject *)x);							// dropper free function
}
