/* -----------------------------------------------------------------------------
 *
 *	threshmod~
 *		modulates signals above a complimentary positive/negative threshold
 *		y[n] = t + ( t - x[n] )
 *
 *	2007 . Cooper Baker . www.cooperbaker.com
 *
 *----------------------------------------------------------------------------*/

#include <Carbon/Carbon.h>
#include <MaxAPI/MaxAPI.h>
#include <MaxAudioAPI/MaxAudioAPI.h>

void *threshmod_class;									// pointer to this 'class'

// threshmod~ structure ------------------------------------------------------ 
typedef struct _threshmod								// threshmod data structure
{ 
    t_pxobject x_obj;									// every external needs this; t_pxobject struct contains object info
	Float32	carrier;
	Float32 modulator;
	Float32 threshpos;
	Float32 threshneg;
	Float32 depth;
	Float32 saturation;
} t_threshmod; 

// prototypes ------------------------------------------------------------------ 
void *threshmod_new(void);								// makes a new threshmod~ object
void threshmod_free(t_threshmod *x);					// deletes the threshmod~ object
void threshmod_dsp(t_threshmod *x, t_signal **sp, short *count);	// sets up dsp for threshmod
t_int *threshmod_perform(t_int *w);					// the dsp perform function
void threshmod_assist(t_threshmod *x, void *b, long m, long a, char *s);
														// inlet / outlet assistance
// main ------------------------------------------------------------------------
void main(void) 
{ 
	setup((t_messlist **)&threshmod_class, (method)threshmod_new, (method)threshmod_free, (short)sizeof(t_threshmod), 0L, 0); 
	addmess((method)threshmod_dsp, "dsp", A_CANT, 0);	// add dsp method
	addmess((method)threshmod_assist, "assist", A_CANT, 0);// assist method 
	dsp_initclass();									// init the class
	post( "" );
	post( "cb.threshmod~ :: modulates signals above a variable threshold" );
	post( "2007 . Cooper Baker . www.cooperbaker.com" );
	post( "" );
}														// post after init completes

// threshmod_dsp --------------------------------------------------------------------
void threshmod_dsp(t_threshmod *x, t_signal **sp, short *count)	// threshmod_dsp(*data struct, **signal struct, signal count )
{ 
	dsp_add( threshmod_perform, 7, sp[0]->s_vec, sp[1]->s_vec, sp[2]->s_vec, sp[3]->s_vec, sp[4]->s_vec, sp[0]->s_n, x );
}							  // 7 = number of values passed to threshmod_perform

// threshmod_perform ---------------------------------------------------------------- 
t_int *threshmod_perform(t_int *w)						// threshmod_perform runs once per signal vector
{ 
	t_float *in1	= (t_float *)(w[1]);				// inlet 1 signal vector location
	t_float *in2	= (t_float *)(w[2]);				// inlet 2 signal vector location
	t_float *in3	= (t_float *)(w[3]);				// inlet 3 signal vector location
	t_float *in4	= (t_float *)(w[4]);				// inlet 4 signal vector location	
	t_float *out	= (t_float *)(w[5]);				// outlet signal vector location
	int n			= (int)(w[6]);						// vector size in samples
	t_threshmod *x	= (t_threshmod *)(w[7]);			// threshmod struct location
	
// DSP BLOCK ///////////////////////////////////////////////////////////////////	
	while ( n-- )										// while a sample remains, do dsp ( n-- is the counter )
	{  
		x->carrier		= *in1++;						// *in++/*out++ should only be done once per while loop cycle
		x->modulator		= *in2++;
		x->threshpos	= *in3++;						// *in++/*out++ should only be done once per while loop cycle
		x->depth		= *in4++;
		
		x->threshpos	= ( x->threshpos < 0 ) ? 0 : x->threshpos;	// clip threshold below 0
		x->threshpos	= ( x->threshpos > 1 ) ? 1 : x->threshpos;	// clip threshold above 1
		x->threshneg	= x->threshpos * -1;			// calculate negative thresh

// out = thresh + ( in * ( modulator * depth * ( ( in - thresh ) / ( 1 - thresh ) ) ) )

		if( x->carrier > x->threshpos )					// if sample is above pos thresh
		{	
			x->saturation = ( x->carrier - x->threshpos ) / ( 1.0 - x->threshpos ); 
			x->carrier = ( x->threshpos + ( x->carrier - x->threshpos) ) + ( x->carrier * ( x->modulator * x->depth * x->saturation ));
			
			goto exit;									// exit
		}
		if( x->carrier < x->threshneg )					// if sample is below neg thresh
		{
			x->saturation = ( ( x->carrier - x->threshneg ) / ( -1.0 - x->threshneg ) ); 
			x->carrier = ( x->threshneg - ( x->threshneg - x->carrier ) ) + ( x->carrier * ( x->modulator * x->depth * x->saturation));
			goto exit;									// exit
		}
		
		exit :											// the exit
		*out++ = x->carrier;
	}
// END DSP BLOCK ///////////////////////////////////////////////////////////////	
	
	return (w + 8);										// always return w plus one more than the 2nd argument in dsp_add() 
} 

// threshmod_assist -----------------------------------------------------------------
void threshmod_assist(t_threshmod *x, void *b, long m, long a, char *s)
{
	switch (m)											// inlet / outlet switch
	{
		case 1:											// 1 = inlets
			switch (a)									// inlet number switch
			{
				case 0:									// inlet 0
					sprintf(s, "(signal) threshmod~ carrier");
					break;								// assist message
				case 1:									// inlet 1
					sprintf(s, "(signal) threshmod~ modulator");
					break;								// assist message
				case 2:
					sprintf(s, "(signal) threshmod~ threshold" );
					break;
				case 3:
					sprintf(s, "(signal) threshmod~ depth" );
				default:
					break;
			}
			break;
		case 2:											// 2 = outlets
			switch (a)									// outlet number switch
			{
				case 0:									// outlet 0
					sprintf(s, "(signal) threshmod~ output");
					break;								// assist message
				default:
					break;
			}
			break;
	}
}

// threshmod_new() ------------------------------------------------------------------ 
void *threshmod_new(void)								// threshmod~ 'constructor'
{ 
	t_threshmod *x = (t_threshmod *)newobject(threshmod_class);		// register the object with threshmod and save its address as *x
	dsp_setup((t_pxobject *)x, 4);						// create inlets
	outlet_new((t_pxobject *)x, "signal");				// create outlet 
	return (x);											// where am i?
} 

// threshmod_free() -----------------------------------------------------------------
void threshmod_free(t_threshmod *x)						// threshmod~ 'destructor'
{
	dsp_free((t_pxobject *)x);							// threshmod free function
}
