/* -----------------------------------------------------------------------------
 *
 *	threshspike~
 *		turns signals into single sample spikes at each wave peak above the threshold
 *		
 *
 *	2007 . Cooper Baker . www.cooperbaker.com
 *
 *----------------------------------------------------------------------------*/

#include <Carbon/Carbon.h>
#include <MaxAPI/MaxAPI.h>
#include <MaxAudioAPI/MaxAudioAPI.h>

void *threshspike_class;									// pointer to this 'class'

// threshspike~ structure ------------------------------------------------------ 
typedef struct _threshspike									// threshspike data structure
{ 
    t_pxobject x_obj;										// every max external needs this; t_pxobject struct contains object info
	Float32		sample[3];									// sample memory for analysis
	Float32		spike;										// output sample
	Float32		thresh;										// positive threshold
	Float32		negthresh;									// negative threshold
} t_threshspike; 

// prototypes ------------------------------------------------------------------ 
void *threshspike_new(void);								// makes a new threshspike~ object
void threshspike_free(t_threshspike *x);					// deletes the threshspike~ object
void threshspike_dsp(t_threshspike *x, t_signal **sp, short *count); // sets up dsp for threshspike
t_int *threshspike_perform(t_int *w);						// the dsp perform function
void threshspike_assist(t_threshspike *x, void *b, long m, long a, char *s);
															// inlet / outlet assistance
// main ------------------------------------------------------------------------
void main(void) 
{ 
	setup((t_messlist **)&threshspike_class, (method)threshspike_new, (method)threshspike_free, (short)sizeof(t_threshspike), 0L, 0); 
	addmess((method)threshspike_dsp, "dsp", A_CANT, 0);		// add dsp method
	addmess((method)threshspike_assist, "assist", A_CANT, 0); // assist method 
	dsp_initclass();	
	post( "" );								
	post( "cb.threshspike~ :: creates spikes at peaks above a variable threshold" );
	post( "2007 . Cooper Baker . www.cooperbaker.com" );
	post( "" );
}															// post after init completes

// threshspike_dsp --------------------------------------------------------------------
void threshspike_dsp(t_threshspike *x, t_signal **sp, short *count)
{															// threshspike_dsp(*data struct, **signal struct, signal count )
	dsp_add( threshspike_perform, 5, sp[0]->s_vec, sp[1]->s_vec, sp[1]->s_vec, sp[0]->s_n, x );
}							   // 5 = number of values passed to threshspike_perform

// threshspike_perform ---------------------------------------------------------------- 
t_int *threshspike_perform(t_int *w)						// threshspike_perform runs once per signal vector
{ 
	t_float *in1	= (t_float *)(w[1]);					// inlet signal vector location
	t_float *in2	= (t_float *)(w[2]);					// inlet signal vector location
	t_float *out	= (t_float *)(w[3]);					// outlet signal vector location
	int n			= (int)(w[4]);							// vector size in samples
	t_threshspike *x= (t_threshspike *)(w[5]);				// threshspike struct location
	
	int i;
// DSP BLOCK ///////////////////////////////////////////////////////////////////	
	while ( n-- )											// while a sample remains, do dsp ( n-- is the counter )
	{  
		x->sample[2] = x->sample[1];
		x->sample[1] = x->sample[0];
		x->sample[0] = *in1++;								// *in++/*out++ should only be done once per while loop cycle
		x->spike = x->sample[1];
		
		x->thresh = *in2++;
		x->thresh = ( x->thresh < 0 ) ? x->thresh * -1 : x->thresh;
		x->negthresh = x->thresh * -1;
		
		if( x->sample[1] > x->thresh )
		{
			if( ( x->sample[0] < x->sample[1] ) && ( x->sample[1] > x->sample[2] ) )
			{
				x->spike = x->sample[1];
			}
			else
			{
				x->spike = x->thresh;
			}
		}
		if( x->sample[1] < x->negthresh )
		{
			if( ( x->sample[0] > x->sample[1] ) && ( x->sample[1] < x->sample[2] ) )
			{
				x->spike = x->sample[1];
			}
			else
			{
				x->spike = x->negthresh;
			}
		}
		*out++ = x->spike;
	}
// END DSP BLOCK ///////////////////////////////////////////////////////////////	
	
	return (w + 6);											// always return w plus one more than the 2nd argument in dsp_add() 
} 

// threshspike_assist -----------------------------------------------------------------
void threshspike_assist(t_threshspike *x, void *b, long m, long a, char *s)
{
	switch (m)												// inlet / outlet switch
	{
		case 1:												// 1 = inlets
			switch (a)										// inlet number switch
			{
				case 0:										// inlet 0
					sprintf(s, "(signal) input");
					break;									// assist message
				case 1:										// inlet 0
					sprintf(s, "(signal) threshold");
					break;									// assist message
				default:
					break;
			}
			break;
		case 2:												// 2 = outlets
			switch (a)										// outlet number switch
			{
				case 0:										// outlet 0
					sprintf(s, "(signal) output");
					break;									// assist message
				default:
					break;
			}
			break;
	}
}

// threshspike_new() ------------------------------------------------------------------ 
void *threshspike_new(void)									// threshspike~ 'constructor'
{ 
	t_threshspike *x = (t_threshspike *)newobject(threshspike_class); // register the object with threshspike and save its address as *x
	dsp_setup((t_pxobject *)x, 2);							// create left inlet 
	outlet_new((t_pxobject *)x, "signal");					// create left outlet 
	
	x->sample[0] = 0;										// init x->samples
	x->sample[1] = 0;
	x->sample[2] = 0;
	
	return (x);												// where am i?
} 

// threshspike_free() -----------------------------------------------------------------
void threshspike_free(t_threshspike *x)						// threshspike~ 'destructor'
{
	dsp_free((t_pxobject *)x);								// threshspike free function
}
