/* -----------------------------------------------------------------------------
 *
 *	verb~
 *		physical model of three parallel surfaces
 *		
 *
 *	2007 . Cooper Baker . www.cooperbaker.com
 *
 *----------------------------------------------------------------------------*/

#include <Carbon/Carbon.h>
#include <MaxAPI/MaxAPI.h>
#include <MaxAudioAPI/MaxAudioAPI.h>

void *verb_class;										// global pointer to verb

typedef struct _verb									// verb struct
{
	t_pxobject v_obj;									// v_obj of type t_pxobject
	float v_sr;											// object sample rate
	float v_table_max;									// max table sample size
	long v_x_samps;										// samples per delay line
	long v_y_samps;										//
	long v_z_samps;										//
	long v_x_write;		 								// write pointers
	long v_y_write;										//
	long v_z_write;										//
	long v_x1_read;										// read pointers
	long v_x2_read;										//
	long v_y1_read;										//
	long v_y2_read;										//
	long v_z1_read;										//	
	long v_z2_read;										//
	float *v_x_table;									// delay tables
	float *v_y_table;									//	
	float *v_z_table;									//
	float v_x_coeff;									// position coefficients
	float v_y_coeff;									//
	float v_z_coeff;									//
	float v_a_coeff;									// absorbtion coefficient
} t_verb;

// Prototypes --------------------------------------------
void *verb_new(void);
t_int *verb_perform(t_int *w);
void verb_dsp(t_verb *x, t_signal **sp, short *count);
void verb_free(t_verb *x);
void verb_x_dimension(t_verb *x, float size);
void verb_y_dimension(t_verb *x, float size);
void verb_z_dimension(t_verb *x, float size);
void verb_x_position(t_verb *x, float coeff);
void verb_y_position(t_verb *x, float coeff);
void verb_z_position(t_verb *x, float coeff);
void verb_a_coefficient(t_verb *x, float coeff);
void verb_assist(t_verb *x, void *b, long m, long a, char *s);
// Prototypes --------------------------------------------

void main(void)
{
	setup((t_messlist**)&verb_class, (method)verb_new, (method)verb_free, (short)sizeof(t_verb), 0L, 0);
	addmess((method)verb_dsp, "dsp", A_CANT, 0);		// dsp method
	addmess((method)verb_assist, "assist", A_CANT, 0);	// assist method 
	addftx((method)verb_x_dimension, 1);				// float inlet methods
	addftx((method)verb_y_dimension, 2);				//
	addftx((method)verb_z_dimension, 3);				//
	addftx((method)verb_x_position, 4);				//
	addftx((method)verb_y_position, 5);				//
	addftx((method)verb_z_position, 6);				//
	addftx((method)verb_a_coefficient, 7);				//
	dsp_initclass();
	post( "cb.verb~ :: physical model of three parallel surfaces" );
	post( "2007 . Cooper Baker . www.cooperbaker.com" );
	post( "" );
}

t_int *verb_perform(t_int *w)
{
	t_float *in = (t_float*)(w[1]);						// input
	t_float *out = (t_float*)(w[2]);					// output
	t_verb *x = (t_verb*)(w[3]);						// verb struct location
	long n = w[4];										// signal vector size
	float in_samp = 0;									// input sample temp
	float x1 = 0;										// delay line output temps
	float x2 = 0;										//
	float x3 = 0;										//
	float y1 = 0;										//
	float y2 = 0;										//	
	float y3 = 0;										//
	float z1 = 0;										//
	float z2 = 0;										//
	float z3 = 0;										//
	float room = 0;										// summed delay lines temp
	while (n--)											// calculation loop
	{
		in_samp = *in++;								// store input sample
// X dimension 											// calculate X read pointers
		x->v_x1_read = (long)(x->v_x_write + (x->v_x_coeff * x->v_x_samps));
		x->v_x2_read = (long)(x->v_x_write + ((1 - x->v_x_coeff) * x->v_x_samps));
		while (x->v_x1_read > x->v_x_samps)			// wrap X read pointers
			x->v_x1_read -= x->v_x_samps;				//
		while (x->v_x1_read < 0)						//
			x->v_x1_read += x->v_x_samps;				//	
		while (x->v_x2_read > x->v_x_samps)			//	
			x->v_x2_read -= x->v_x_samps;				//
		while (x->v_x2_read < 0)						//
			x->v_x2_read += x->v_x_samps;				//
		x1 = *(x->v_x_table + x->v_x1_read);			// X delay line tap 1
		x2 = *(x->v_x_table + x->v_x2_read);			// X delay line tap 2
		x3 = ((x1 + x2) * 0.7071);						// X delay line tap sum * -3dB
// Y dimension											// calculate Y read pointers
		x->v_y1_read = (long)(x->v_y_write + (x->v_y_coeff * x->v_y_samps));
		x->v_y2_read = (long)(x->v_y_write + ((1 - x->v_y_coeff) * x->v_y_samps));		
		while (x->v_y1_read > x->v_y_samps)			// wrap Y read pointers
			x->v_y1_read -= x->v_y_samps;				//
		while (x->v_y1_read < 0)						//
			x->v_y1_read += x->v_y_samps;				//
		while (x->v_y2_read > x->v_y_samps)			//	
			x->v_y2_read -= x->v_y_samps;				//
		while (x->v_y2_read < 0)						//
			x->v_y2_read += x->v_y_samps;				//
		y1 = *(x->v_y_table + x->v_y1_read);			// Y delay line tap 1
		y2 = *(x->v_y_table + x->v_y2_read);			// Y delay line tap 2
		y3 = ((y1 + y2) * 0.7071);						// Y delay line tap sum * -3dB
// Z dimension											// calculate Z read pointers
		x->v_z1_read = (long)(x->v_z_write + (x->v_z_coeff * x->v_z_samps));
		x->v_z2_read = (long)(x->v_z_write + ((1 - x->v_z_coeff) * x->v_z_samps));
		while (x->v_z1_read > x->v_z_samps)			// wrap Z read pointers
			x->v_z1_read -= x->v_z_samps;				//
		while (x->v_z1_read < 0)						//
			x->v_z1_read += x->v_z_samps;				//
		while (x->v_z2_read > x->v_z_samps)			//
			x->v_z2_read -= x->v_z_samps;				//
		while (x->v_z2_read < 0)						//
			x->v_z2_read += x->v_z_samps;				//
		z1 = *(x->v_z_table + x->v_z1_read);			// Z delay line tap 1
		z2 = *(x->v_z_table + x->v_z2_read);			// Z delay line tap 2
		z3 = ((z1 + z2) * 0.7071);						// Z delay line tap sum * -3dB
		room = ((((x3 + y3) * 0.7071) + z3) * 0.7071);	// sum all delay lines 
														// fill tables with input and feedback		
		*(x->v_x_table + x->v_x_write) = (in_samp + x3) * x->v_a_coeff * 0.7071;
		*(x->v_y_table + x->v_y_write) = (in_samp + y3) * x->v_a_coeff * 0.7071;
		*(x->v_z_table + x->v_z_write) = (in_samp + z3) * x->v_a_coeff * 0.7071;
		*out++ = room;									// output the verb	
		x->v_x_write++;									// increment write pointers
		x->v_y_write++;									//
		x->v_z_write++;									//
		while (x->v_x_write > x->v_x_samps)			// wrap X write pointers
			x->v_x_write -= x->v_x_samps;				//
		while (x->v_x_write < 0)						//
			x->v_x_write += x->v_x_samps;				//
		while (x->v_y_write > x->v_y_samps)			// wrap Y write pointers
			x->v_y_write -= x->v_y_samps;				// 
		while (x->v_y_write < 0)						//
			x->v_y_write += x->v_y_samps;				//
		while (x->v_z_write > x->v_z_samps)			// wrap Z write pointers
			x->v_z_write -= x->v_z_samps;				//
		while (x->v_z_write < 0)						//
			x->v_z_write += x->v_z_samps;				//
	}
	return(w+5);										// return signal struct location + # vars + 1
}

void verb_x_dimension(t_verb *x, float size)
{
	if (size < 1)										// constrain X size
		size = 1;										//
	if (size > 5280)									//
		size = 5280;									//
														// convert feet to samples and store
	x->v_x_samps = (2 * (size * 0.0009090909090909091) * 1000) * (x->v_sr * (0.001));	
}

void verb_y_dimension(t_verb *x, float size)
{
	if (size < 1)										// constrain Z size
		size = 1;										//
	if (size > 5280)									//
		size = 5280;									//
														// convert feet to samples and store
	x->v_y_samps = (2 * (size * 0.0009090909090909091) * 1000) * (x->v_sr * (0.001));	
}

void verb_z_dimension(t_verb *x, float size)
{
	if (size < 1)										// constrain Y size
		size = 1;										//
	if (size > 5280)									//
		size = 5280;									//
														// convert feet to samples and store
	x->v_z_samps = (2 * (size * 0.0009090909090909091) * 1000) * (x->v_sr * (0.001));
}

void verb_x_position(t_verb *x, float coeff)
{
	if (coeff < 0)										// constrain X position coefficient
		coeff = 0;										//
	if (coeff > 1)										//
		coeff = 1;										//
	x->v_x_coeff = coeff;								// store position coefficient
}

void verb_y_position(t_verb *x, float coeff)
{
	if (coeff < 0)										// constrain Y position coefficient
		coeff = 0;										//
	if (coeff > 1)										//
		coeff = 1;										//
	x->v_y_coeff = coeff;								// store position coefficient
}

void verb_z_position(t_verb *x, float coeff)
{
	if (coeff < 0)										// constrain Z position coefficient
		coeff = 0;										//
	if (coeff > 1)										//
		coeff = 1;										//
	x->v_z_coeff = coeff;								// store position coefficient
}

void verb_a_coefficient(t_verb *x, float coeff)
{
	if (coeff < 0)										// constrain absorbtion coefficient
		coeff = 0;										//
	if (coeff > 1)										//
		coeff = 1;										//
	x->v_a_coeff = 1 - coeff;							// invert and store absorbtion coefficient
}

void *verb_new(void)
{
	t_verb *x = (t_verb*)newobject(verb_class);		// Make verb~ object	
	dsp_setup((t_pxobject *)x, 1);						// Initialize instance, 1 inlet
	outlet_new((t_pxobject *)x, "signal");				// Signal outlet
	floatin(x, 7);										// Make control inlets
	floatin(x, 6);										//
	floatin(x, 5);										//
	floatin(x, 4);										//
	floatin(x, 3);										//
	floatin(x, 2);										//
	floatin(x, 1);										//
	x->v_x_samps = 1;									// Initializations
	x->v_y_samps = 1;									//
	x->v_z_samps = 1;									//	
	x->v_table_max = 460801;							//
	x->v_x_table = (float*)sysmem_newptrclear((long)x->v_table_max*sizeof(float));
	x->v_y_table = (float*)sysmem_newptrclear((long)x->v_table_max*sizeof(float));
	x->v_z_table = (float*)sysmem_newptrclear((long)x->v_table_max*sizeof(float));
	x->v_x_coeff = 0.2;									//
	x->v_y_coeff = 0.4;									//
	x->v_z_coeff = 0.7;									//
	x->v_a_coeff = 0.5;									//	
	x->v_x_write = 0;									//
	x->v_y_write = 0;									//
	x->v_z_write = 0;									//
	x->v_x1_read = 0;									//
	x->v_x2_read = 0;									//
	x->v_y1_read = 0;									//
	x->v_y2_read = 0;									//
	x->v_z1_read = 0;									//
	x->v_z2_read = 0;									//
	x->v_sr = 44100;									//	
	verb_x_dimension(x, 30);							// Initialize room dimensions
	verb_y_dimension(x, 42);							//
	verb_z_dimension(x, 51);							//
	return x;											// Where am I?
}

void verb_dsp(t_verb *x, t_signal **sp, short *count)
{
	x->v_sr = sp[0]->s_sr;
	dsp_add(verb_perform, 4, sp[0]->s_vec, sp[1]->s_vec, x, sp[0]->s_n);
}

void verb_free(t_verb *x)							
{
	if (x->v_x_table != 0)								// free delay line memory
		sysmem_freeptr(x->v_x_table);					//
	x->v_x_table = 0;									//
	if (x->v_y_table != 0)								//
		sysmem_freeptr(x->v_y_table);					//	
	x->v_y_table = 0;									//
	if (x->v_z_table != 0)								//
		sysmem_freeptr(x->v_z_table);					//	
	x->v_z_table = 0;									//
	dsp_free((t_pxobject *)x);							// msp free function
}

void verb_assist(t_verb *x, void *b, long m, long a, char *s)
{
	switch (m)											// inlet and outlet assistance
	{
		case 1:
			switch (a)
			{
				case 0:
					sprintf(s, "(Signal) Input");
					break;
				case 1:
					sprintf(s, "(Float) X Size (feet, 0 - 5280)");
					break;
				case 2:
					sprintf(s, "(Float) Y Size (feet, 0 - 5280)");
					break;
				case 3:
					sprintf(s, "(Float) Z Size (feet, 0 - 5280)");
					break;
				case 4:
					sprintf(s, "(Float) X Position Coefficient (0 - 1)");
					break;
				case 5:
					sprintf(s, "(Float) Y Position Coefficient (0 - 1)");
					break;
				case 6:
					sprintf(s, "(Float) Z Position Coefficient (0 - 1)");
					break;
				case 7:
					sprintf(s, "(Float) Absorbtion Coefficient (0 - 1)");
					break;		
			}
		break;
		case 2:
			switch (a)
			{
				case 0:
					sprintf(s, "(Signal) Output");
					break;
			}
		break;
	}
}


